import os
import subprocess
from abc import ABC, abstractmethod
from collections import namedtuple
from dataclasses import dataclass
from typing import Any, Iterable, Optional

import toml
import typer
from git import Repo
from git.exc import GitCommandError

VersionBump = namedtuple("VersionBump", ["previous_version", "new_version"])


@dataclass(kw_only=True)
class Result:
    success: bool
    messages: Optional[Iterable[str]] = None
    return_value: Any = None


@dataclass(kw_only=True)
class StepMixin:
    short_name: str
    description: str
    project_dir: Optional[str] = None


class Step(ABC, StepMixin):
    @abstractmethod
    def run(self) -> Result:
        raise NotImplementedError  # pragma: no cover


class GitStep(Step):
    def __post_init__(self):
        self.repo = Repo(self.project_dir)  # pragma: no cover


class DynamicVersionDescriptionGitStep(GitStep):
    _new_version: str
    prefix: str

    @property
    def new_version(self) -> str:
        return self._new_version

    @new_version.setter
    def new_version(self, new_v: str):
        self.description += f" [blue]{self.prefix}{new_v}[/blue]"
        self._new_version = new_v


@dataclass(kw_only=True)
class CheckPyProjectStep(Step):
    short_name: str = "check_pyproject_toml"
    description: str = "Make sure you have a (non-empty) pyproject.toml"

    def run(self) -> Result:
        success = (
            os.path.isfile("pyproject.toml") and os.path.getsize("pyproject.toml") > 0
        )
        return Result(success=success)


@dataclass(kw_only=True)
class CheckOnMainStep(GitStep):
    short_name: str = "check_on_main"
    description: str = "Make sure you're on main/master branch"

    def run(self) -> Result:
        current_branch = self.repo.active_branch.name
        success = current_branch == "main" or current_branch == "master"
        return Result(success=success)


@dataclass(kw_only=True)
class CheckCleanGitStep(GitStep):
    short_name: str = "check_clean_git"
    description: str = "Make sure git is clean"
    allow_untracked_files: bool = False

    def run(self) -> Result:
        if self.repo.is_dirty(untracked_files=self.allow_untracked_files):
            return Result(success=False, messages=["You have an unclean working tree!"])
        res = self.repo.git.status("--branch", "--porcelain")
        lines = str.splitlines(res)
        if len(lines) > 0 and "[ahead " in lines[0]:
            return Result(success=False, messages=["Local and remote not synced!"])
        return Result(success=True)


@dataclass(kw_only=True)
class RunTestsStep(Step):
    short_name: str = "run_tests"
    description: str = "Run test-suite"

    def run(self) -> Result:
        success = subprocess.run(["nox", "-r"]).returncode == 0
        return Result(success=success)


@dataclass(kw_only=True)
class CheckChangelogStep(GitStep):
    short_name: str = "check_changelog"
    description: str = "Make sure changelog was updated"

    def run(self) -> Result:
        paths = [
            "CHANGELOG",
            "CHANGELOG.md",
            "CHANGELOG.rst",
        ]
        cl_path = None
        for p in paths:
            if os.path.isfile(p):
                cl_path = p
                break
        if cl_path is None:
            return Result(
                success=False,
                messages=["Could not find Changelog in paths:", f"{paths}"],
            )
        current_version: str = toml.load("pyproject.toml")["tool"]["poetry"]["version"]
        if current_version == "0.1.0":
            if os.path.getsize(cl_path) > 0:
                # there is a changelog file and it contains something
                return Result(success=True)
            else:
                return Result(
                    success=False, messages=[f"Changelog file '{cl_path}' empty"]
                )
        # check if there is a diff
        res = self.repo.git.diff(current_version, "--", cl_path)
        success = len(res) > 0
        return Result(success=success)


@dataclass(kw_only=True)
class BumpVersionStep(Step):
    bump_rule: str
    short_name: str = "bumpversion"
    description: str = "Bump version"

    def run(self, dry_run: bool = False) -> Result:
        command = ["poetry", "version", self.bump_rule]
        if dry_run:
            command.append("--dry-run")
        res = subprocess.run(command, capture_output=True)
        if res.returncode == 0:
            bump_message = res.stdout.decode().strip()
            previous_version = bump_message.split(" ")[3]
            new_version = bump_message.split(" ")[5]
            part_message = f"version from [yellow]{previous_version}[/yellow] to [blue]{new_version}[/blue]"
            if dry_run:
                return Result(
                    success=True,
                    messages=[f"Would bump {part_message}"],
                    return_value=VersionBump(
                        previous_version=previous_version, new_version=new_version
                    ),
                )
            return Result(success=True, messages=[f"Bumped {part_message}"])
        return Result(success=False, messages=str.splitlines(res.stderr.decode()))


@dataclass(kw_only=True)
class CommitVersionBumpStep(GitStep):
    short_name: str = "commitbump"
    description: str = "Commit version bump"
    commit_message: str = "Bumped version"

    def run(self) -> Result:
        try:
            self.repo.git.add("pyproject.toml")
            self.repo.git.commit("-m", self.commit_message)
        except GitCommandError as e:
            return Result(success=False, messages=[e.stderr.strip()])
        return Result(success=True)


@dataclass(kw_only=True)
class GitTagStep(DynamicVersionDescriptionGitStep):
    short_name: str = "gittag"
    description: str = "Create git tag"
    prefix: str = "v"

    def run(self) -> Result:
        if not hasattr(self, "_new_version"):
            return Result(
                success=False, messages=["Can't tag if no new version is provided"]
            )
        try:
            self.repo.git.tag(f"{self.prefix}{self.new_version}")
        except GitCommandError as e:
            return Result(success=False, messages=[e.stderr.strip()])
        return Result(success=True)


@dataclass(kw_only=True)
class PushTagStep(DynamicVersionDescriptionGitStep):
    short_name: str = "pushtag"
    description: str = "Push git tag"
    remote: str = "origin"
    prefix: str = "v"

    def run(self) -> Result:
        if self.new_version is None:
            return Result(
                success=False, messages=["Can't tag if no new version is provided"]
            )
        try:
            self.repo.git.push(self.remote, f"{self.prefix}{self.new_version}")
        except GitCommandError as e:
            return Result(success=False, messages=[e.stderr.strip()])
        return Result(success=True)


@dataclass(kw_only=True)
class PoetryPublishStep(Step):
    short_name: str = "publish"
    description: str = "Build and publish with poetry"

    def run(self) -> Result:
        command = ["poetry", "publish", "--build"]
        if os.getenv("PYPI_TOKEN"):
            token = os.getenv("PYPI_TOKEN")
            assert token  # for mypy
            command.extend(["--username", "__token__", "--password", token])
        else:
            user_name = typer.prompt("Please enter your PyPI username")
            password = typer.prompt("Please enter your PyPI password", hide_input=True)
            command.extend(["--username", user_name, "--password", password])
        res = subprocess.run(command)
        if res.returncode != 0:
            return Result(success=False, messages=[res.stderr.decode().strip()])
        return Result(success=True, messages=["Build and published successfully!"])


@dataclass(kw_only=True)
class GithubReleaseReminderStep(Step):
    short_name: str = "githubreleasreminder"
    description: str = "Remind you to upload build as github release"

    def run(self) -> Result:
        gh_released = typer.confirm(
            "Please create a github release now! Did you do it?"
        )
        if not gh_released:
            return Result(
                success=False,
                messages=["You did not want to create a github release..."],
            )
        return Result(success=True)



from dataclasses import dataclass
from typing import Iterable, Optional

from rich import print as rprint


@dataclass
class ReciteConsole:
    prefix: str = "recite >"
    base_color: str = "white"
    bad_color: str = "red"
    good_color: str = "green"
    bad_glyph: str = "✘ "
    good_glyph: str = "✓ "

    def print_message(
        self,
        message: str,
        color: str = "white",
        glyph: str = "",
        number_str: str = "",
        indent_count: int = 0,
        indent_char: str = "*",
    ):
        if color.lower() == "bad":
            color = self.bad_color
        elif color.lower() == "good":
            color = self.good_color
        indent_str = ""
        if indent_count > 0:
            indent_str = "\t" * indent_count
            indent_str += indent_char + " "
        rprint(
            f"{self.prefix} {number_str}[{color}]{indent_str}{glyph}{message}[/{color}]"
        )

    def print_multiple_messages(
        self,
        messages: Iterable[str],
        color: str = "white",
        indent_count: int = 0,
        indent_char: str = "*",
    ):
        for msg_n, msg in enumerate(messages):
            if msg_n > 0:
                indent_char = " "
            self.print_message(
                message=msg,
                indent_count=indent_count,
                indent_char=indent_char,
                color=color,
            )

    def print_success(self, message: str, number: Optional[int] = None):
        number_str = ""
        if number:
            number_str = f"{number}: "
        self.print_message(
            color=self.good_color,
            glyph=self.good_glyph,
            number_str=number_str,
            message=message,
        )

    def print_failure(self, message: str, number: Optional[int] = None):
        number_str = ""
        if number:
            number_str = f"{number}: "
        self.print_message(
            color=self.bad_color,
            glyph=self.bad_glyph,
            number_str=number_str,
            message=message,
        )

    def print_error(self, message: str, indent_count: int):
        self.print_message(
            message=message, indent_count=indent_count, color=self.bad_color
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['recite']

package_data = \
{'': ['*']}

modules = \
['py']
install_requires = \
['gitpython>=3.1.30,<4.0.0',
 'rich>=13.1.0,<14.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer>=0.7.0,<0.8.0']

extras_require = \
{'docs': ['mkdocs>=1.4.2,<2.0.0', 'mkdocs-material>=9.0.9,<10.0.0']}

entry_points = \
{'console_scripts': ['recite = recite.main:app']}

setup_kwargs = {
    'name': 'recite',
    'version': '0.1.1',
    'description': 'Publish your poetry-based projects, without missing important steps',
    'long_description': '<p align="center">\n<img src="https://github.com/dobraczka/recite/raw/main/docs/assets/logo.png" alt="recite logo", width=200/>\n</p>\n<h2 align="center"> recite</h2>\n\n\n<p align="center">\n<a href="https://github.com/dobraczka/recite/actions/workflows/main.yml"><img alt="Actions Status" src="https://github.com/dobraczka/recite/actions/workflows/main.yml/badge.svg?branch=main"></a>\n<a href=\'https://recite.readthedocs.io/en/latest/?badge=latest\'><img src=\'https://readthedocs.org/projects/recite/badge/?version=latest\' alt=\'Documentation Status\' /></a>\n<a href="https://codecov.io/gh/dobraczka/recite"><img src="https://codecov.io/gh/dobraczka/recite/branch/main/graph/badge.svg?token=TCMKS9U0MH"/></a>\n<a href="https://pypi.org/project/recite"/><img alt="Stable python versions" src="https://img.shields.io/pypi/pyversions/recite"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\n`recite` intends to make releasing [poetry](https://python-poetry.org/)-based libraries easier and avoid missing important steps (e.g. adapting the changelog (because [you should keep one](https://keepachangelog.com/))).\n\n```console\n$ recite patch\n\nrecite > 👀 Checking everything to make sure you are ready to release 👀\nrecite > 1: ✓ Make sure you have a (non-empty) pyproject.toml\nrecite > 2: ✓ Make sure you\'re on main/master branch\nrecite > 3: ✓ Make sure git is clean\nrecite > 4: ✓ Run test-suite\nrecite > 5: ✓ Make sure changelog was updated\nrecite > 🤓 Everything looks perfect! 🤓\nrecite > I will perform the following steps:\nrecite >        * Would bump version from 0.1.0 to 0.1.1\nrecite >        * Commit version bump\nrecite >        * Create git tag 0.1.1\nrecite >        * Push git tag 0.1.1\nrecite >        * Remind you to upload build as github release\nDo you want to proceed? [y/N]: y\nrecite > ✨ Performing release ✨\nrecite > 1: ✓ Bump version\nrecite >        * Bumped version from 0.1.0 to 0.1.1\nrecite > 2: ✓ Commit version bump\nrecite > 3: ✓ Create git tag 0.1.1\nrecite > 4: ✓ Push git tag 0.1.1\nrecite > 5: ✓ Build and publish with poetry\nPlease create a github release now! Did you do it? [y/N]: y\nrecite > 6: ✓ Remind you to upload build as github release\nrecite > 🚀 Congrats to your release! 🚀\n```\n\n# Installation\n\nSince `recite` is a python application it is recommended to install it via [pipx](https://pypa.github.io/pipx/):\n```console\n$ pipx install recite\n```\n\nBut you can also install it via pip:\n\n```console\n$ pip install recite\n```\n\n# Why?\n\nPreviously I used a github action to automatically build and publish a new version of a library if a new tag was pushed. However, sometimes I forgot something crucial (e.g. to adapt the changelog). In this case I had to rush to stop the github action before it would publish the release to pypi (where it would lie forever unable to be rectified).\nWith `recite` it is ensured all the necessary checks are in place before any tags are created.\n',
    'author': 'Daniel Obraczka',
    'author_email': 'obraczka@informatik.uni-leipzig.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dobraczka/recite',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
